# Repository path:    $HeadURL: https://90.145.156.222/DMS-Houtribdijk/code/nc-to-thredds/source/spectrum1D.py $
# Last committed:     $Revision: 380303 $; $Author: vvuik $; $Date: 2019-10-21 11:46:46 +0200 (ma, 21 okt 2019) $
#
# -*- coding: utf-8 -*-
"""
Created on Fri Mar 22 14:23:12 2019

@author: vuik

computes variance density

"""

import numpy as np
from scipy import signal
from scipy.fftpack import fft

def nextpow2(x):
    y = np.ceil(np.log2(np.abs(x)))
    return y

def spectrum1D(x,y,fresolution=0.01,detrend=1,overlap=0.5,strict=0,filter='Hann',tolerance=1e-3,correctvar=1):

    # Generates simple spectrum of y(x) with hann filter and welch method
    #
    # [fx vy] = spectrum1D(x,y,varargin) 
    #
    # Input:
    # x = time/distance vector, or scalar measurement interval
    # y = vector of measurements
    # varargin:
    # fresolution = required frequency resolution (0.01)
    # detrend     = logical detrend input signal before FFT (1)
    # overlap     = maximum overlap in Welch method (0.5)
    # figures     = logical output to figure (1)
    # strict      = limit FFT size to power of 2 (0)
    # filter      = filter type: 'Hann','Hamming' or 'none' ('Hann')
    # tolerance   = check for relative error between input and output wave height (1E-3)
    #
    # Output:
    # fx = frequency / wavenumber vector
    # vy = variance density of y
    #
    
    eps = 1e-9 # trick to include endpoints of arange operators
    
    # input checks
    if np.min(np.shape(y))>1 and np.size(np.shape(y))>1:
        raise ValueError('Input y must be a vector')
    else:
        if np.min(np.shape(x))>1 and np.size(np.shape(x))>1:
            raise ValueError('Input x must be a vector')
        elif len(x)>1 and len(x)!=len(y):
            raise ValueError('Input x and input y must have the same length')
        elif len(x)>1 and len(x)==len(y):
            # set to zero base time
            xi = x-x[0]
        elif len(x)==1:
            print('only interval should still be tested')
            xi = np.arange(0,(len(y)-1)*x+eps,x)
            xi = np.reshape(xi,np.size(y))        
    
    # test time/space axis monotonically increasing
    if np.min(np.diff(xi))<0 or np.max(np.diff(xi))-np.min(np.diff(xi))>1E-8: # this is needed for tolerance in rounding numbers
        raise ValueError('Input x must be monotonic increasing')
    
    # Convert to column vectors
    Y=y  # np.reshape(y,np.size(y),1)
    X=xi # np.reshape(xi,np.size(xi),1)
    
    # Time step in time/space axis
    dt = X[1]-X[0]
    
    # Number of data points in the total signal
    N=len(X)
    
    # Number of data points required for fresolution
    Nr = np.ceil((1/fresolution)/dt)
    if strict:
        Nr = 2**nextpow2(Nr)
    
    # Check if there are sufficient data points to acquire the set frequency resolution
    if Nr>N:
        # reset Nr
        Nr = N
        print('The required frequency resolution could not be achieved.')
    
    # Number of Welch repetitions
    Nw = np.ceil((N-Nr)/(overlap*Nr))+1
    
    # Index input arrays for start and end of each Welch repetition
    indend   = np.round(np.linspace(Nr,N,Nw))
    indstart = indend-Nr
    
    # Time and frequency array 
    T  = X[int(indend[0]-1)]
    df = 1/T
    ff = df*np.concatenate((np.arange(0,np.round(Nr/2)+eps,1),np.arange(-1*np.floor(Nr/2)+1,-1+eps,1)), axis=0)
    fx = ff[0:int(np.floor(Nr/2))]
    
    # Storage arrays
    vy=np.zeros(int(np.floor(Nr/2)))
    
    # Detrend input signal
    if detrend:
        Y = signal.detrend(Y)

    
    varpre = np.var(Y)
    # do Welch repetition
    for i in np.arange(0,Nw-1):
        ii = int(i)
        d = Y[int(indstart[ii]):int(indend[ii])]
        if filter=='Hann':
            # Hann filter
            d = d*(0.5*(1-np.cos(2*np.pi*np.arange(0,Nr-1+eps,1)/(Nr-1))))
            varpost = np.var(d)
        elif filter=='Hamming':
            d = d* (0.54-0.46*np.cos(2*np.pi*np.arange(0,Nr-1+eps,1)/(Nr-1)))
            varpost = np.var(d)
        elif filter=='none':
            varpost = varpre
        
        # FFT
        Q = fft(d)
        # Power density
        V = 2*T/Nr**2*np.abs(Q)**2
        # Store in array
        vy = vy + 1/Nw*np.squeeze(V[0:int(np.floor(Nr/2))])
    
    # Restore the total variance
    if correctvar==1:
       vy = (varpre/np.trapz(vy,fx))*vy # trapz(y,x)!
    
    # input/output check
    hrmsin  = 4*np.std(Y)/np.sqrt(2)
    hrmsout = np.sqrt(8*np.trapz(vy,fx)) # trapz(y,x)!
    dif = np.abs(hrmsout-hrmsin)
    if dif > tolerance:
        print('Difference in input and output wave height (' + str(dif) + ') is greater than set tolerance (' + str(tolerance) + ')')
    
    # Negative values in the real component of the result of a complex FFT correspond to 
    # a negative correlation with a cosine waveform (same as a 180 degree phase shift).
    # If you want to check the energy at each frequency, graph the magnitude (sqrt(rere+imim)) of each complex FFT result bin.
    vy = np.abs(vy) 
    
    return fx,vy


    
            
    
    
    
    
    
    
